import subprocess
import sys
import os


class KeyNotFound(Exception):
    pass


class FingerprintDoesNotParse(Exception):
    pass


class KeyLibrary:
    def __init__(self, path, verbose):
        self._path = path
        self._verbose = verbose
        self.keys_by_fingerprint = {}

    def _generate_public_key(self, fname):
        pkey_fname = fname + '.pub'
        if os.access(pkey_fname, os.R_OK):
            return
        print("public key does not exist for private key '%s'" % fname, file=sys.stderr)
        print("attemping to generate; you may be prompted for a passphrase.", file=sys.stderr)
        try:
            public_key = subprocess.check_output(['ssh-keygen', '-y', '-f', fname])
        except subprocess.CalledProcessError:
            return
        with open(pkey_fname, 'wb') as fd:
            fd.write(public_key)
        print("Generated public key successfully.", file=sys.stderr)
        return True

    @classmethod
    def _fingerprint_from_keyinfo(cls, output):
        parts = [s for s in (t.strip() for t in output.split(' ')) if s]
        if len(parts) != 4:
            raise FingerprintDoesNotParse()
            raise
        return parts[1]

    def _scan_key(self, fname, recurse=False):
        try:
            output = subprocess.check_output(['ssh-keygen', '-l', '-f', fname]).decode('utf8')
            try:
                return KeyLibrary._fingerprint_from_keyinfo(output)
            except FingerprintDoesNotParse:
                print("warning: public key fingerprint couldn't be parsed: '%s'" % fname, file=sys.stderr)
                print(output, file=sys.stderr)
        except subprocess.CalledProcessError:
            if not recurse and self._generate_public_key(fname):
                return self._scan_key(fname, recurse=True)

    def scan(self):
        skip = set(('config', 'known_hosts', 'authorized_keys'))
        for dirpath, dirnames, fnames in os.walk(self._path):
            for name, path in ((t, os.path.join(dirpath, t)) for t in fnames):
                if name.startswith('.'):
                    continue
                if name.endswith('.pub'):
                    continue
                if name in skip:
                    continue
                fingerprint = self._scan_key(path)
                if fingerprint is not None:
                    if self._verbose:
                        print("scanned key '%s' fingerprint '%s'" % (os.path.relpath(path, self._path), fingerprint))
                    self.keys_by_fingerprint[fingerprint] = path

    def lookup(self, fingerprint):
        try:
            return self.keys_by_fingerprint[fingerprint]
        except KeyError:
            raise KeyNotFound()



import argparse
import difflib
import os.path
import sys
from tempfile import NamedTemporaryFile
from .engine import SedgeEngine, SedgeException, ConfigOutput
from .keylib import KeyLibrary


def ask_overwrite(fname):
    print(
        "WARNING: `%s' already exists and was not generated by sedge." % fname,
        file=sys.stderr)
    confirm = input("Enter `yes' to overwite: ")
    return confirm == 'yes'


def check_or_confirm_overwrite(fname):
    "returns True if OK to proceed, False otherwise"
    try:
        with open(fname) as fd:
            header = next(fd)
            if header.find(':sedge:') == -1:
                okay = ask_overwrite(fname)
                if okay:
                    backup_file = fname + '.pre-sedge'
                    os.rename(fname, backup_file)
                    print("your previous SSH configuration file has been renamed to:\n%s" % (backup_file), file=sys.stderr)
    except OSError:
        pass
    except StopIteration:
        pass
    return True


def diff_config_changes(before, after):
    def get_data(f):
        with open(f) as fd:
            return fd.read().splitlines(True)
    a = get_data(before)
    b = get_data(after)
    diff_lines = list(difflib.unified_diff(a, b))
    if not diff_lines:
        print("no changes.", file=sys.stderr)
    else:
        print('configuration changes:', file=sys.stderr)
        print(''.join(diff_lines), file=sys.stderr)


def process(args):
    def write_to(out):
        config.output(out)
    library = KeyLibrary(args.key_directory, verbose=args.verbose)
    library.scan()
    with open(args.config_file) as fd:
        config = SedgeEngine(library, fd)
    if args.output_file == '-':
        write_to(ConfigOutput(sys.stdout))
        return
    if not check_or_confirm_overwrite(args.output_file):
        print("Aborting.", file=sys.stderr)
        sys.exit(1)

    tmpf = NamedTemporaryFile(mode='w', dir=os.path.dirname(args.output_file), delete=False)
    try:
        tmpf.file.write('''\
# :sedge:
#
# this configuration generated from `sedge' file:
# %s
#
# do not edit this file manually, edit the source file and re-run `sedge'
#

''' % (args.output_file))
        write_to(ConfigOutput(tmpf.file))
        tmpf.close()
        if args.verbose:
            diff_config_changes(args.output_file, tmpf.name)
        os.rename(tmpf.name, args.output_file)
    except:
        os.unlink(tmpf.name)
        raise


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-k', '--key_directory',
        default=os.path.expanduser('~/.ssh'),
        type=str,
        help='directory to scan for SSH keys')
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='verbose output (including key fingerprints)')
    parser.add_argument(
        '--version', action='store_true',
        help='print version and exit')
    parser.add_argument(
        '-c', '--config-file',
        default=os.path.expanduser('~/.sedge/config'),
        nargs='?')
    parser.add_argument(
        '-o', '--output-file',
        default=os.path.expanduser('~/.ssh/config'),
        nargs='?')
    args = parser.parse_args()
    if args.version:
        import pkg_resources
        version = pkg_resources.require("sedge")[0].version
        print('sedge, version %s' % (version))
        sys.exit(0)
    try:
        process(args)
    except SedgeException as e:
        print('Error: %s' % (e), file=sys.stderr)



from io import StringIO
from .engine import SedgeEngine, Host, ConfigOutput, ParserException, OutputException, SecurityException
from .keylib import KeyLibrary
from nose.tools import eq_, raises


def config_for_text(in_text):
    library = KeyLibrary('/tmp', verbose=False)
    return SedgeEngine(library, StringIO(in_text))


def check_parse_result(in_text, expected_text):
    config = config_for_text(in_text)
    fd = StringIO()
    out = ConfigOutput(fd)
    config.output(out)
    eq_(expected_text, fd.getvalue())


def test_empty_file():
    check_parse_result('', '')


def test_global_option():
    check_parse_result('GlobalOption\n', 'GlobalOption\n')


def test_single_host_stanza():
    check_parse_result('Host blah\n', 'Host = blah\n')


def test_expansion():
    check_parse_result('''
@with i 1 2
Host percival<i>
''', 'Host = percival1\n\nHost = percival2\n')


def test_expansion_long_strings():
    check_parse_result('''
@with i millet corn
Host percival_<i>
''', 'Host = percival_millet\n\nHost = percival_corn\n')


def test_combinatoric_expansion():
    check_parse_result('''
@with i 1 2
@with j 4 5
Host p<i>-<j>
''', 'Host = p1-4\n\nHost = p1-5\n\nHost = p2-4\n\nHost = p2-5\n')


def test_simple_is():
    check_parse_result('''
@HostAttrs a
    SomeOption Yes
Host blah
    @is a
''', 'Host = blah\n    SomeOption = Yes\n')


def test_double_is():
    check_parse_result('''
@HostAttrs a
    SomeOption Yes
@HostAttrs b
    AnotherOption No
Host blah
    @is a
    @is b
''', 'Host = blah\n    SomeOption = Yes\n    AnotherOption = No\n')


def test_via():
    check_parse_result('''
Host blah
    @via gateway
''', 'Host = blah\n    ProxyCommand = ssh gateway nc %h %p 2> /dev/null\n')


def test_include():
    check_parse_result('''
@include https://raw.githubusercontent.com/grahame/sedge/master/ci_data/simple.sedge
''', 'Host = percival\n    HostName = beaking\n    ForwardAgent = yes\n    ForwardX11 = yes\n')


def test_include_strips_root():
    check_parse_result('''
@include https://raw.githubusercontent.com/grahame/sedge/master/ci_data/strip_global.sedge
''', 'Host = percival\n    HostName = beaking\n    ForwardAgent = yes\n    ForwardX11 = yes\n')


def test_include_args():
    check_parse_result('''
@set budgerigar percival
@include https://raw.githubusercontent.com/grahame/sedge/master/ci_data/args.sedge <budgerigar>
''', 'Host = percival\n    HostName = beaking\n')


def check_fingerprint(data, fingerprint):
    determined = KeyLibrary._fingerprint_from_keyinfo(data)
    eq_(determined, fingerprint)


def test_fingerprint_parser():
    check_fingerprint(
        '2048 aa:cb:d2:e2:00:6f:21:b4:fe:39:92:ed:eb:5e:4d:38 grahame@anglachel (RSA)',
        'aa:cb:d2:e2:00:6f:21:b4:fe:39:92:ed:eb:5e:4d:38')


def test_fingerprint_parser_double_space():
    check_fingerprint(
        '2048 aa:cb:d2:e2:00:6f:21:b4:fe:39:92:ed:eb:5e:4d:38  grahame@anglachel (RSA)',
        'aa:cb:d2:e2:00:6f:21:b4:fe:39:92:ed:eb:5e:4d:38')


def check_config_parser(s, expected):
    result = SedgeEngine.parse_config_line(s)
    eq_(result, expected)


def test_parser_noarg():
    check_config_parser('Keyword', ('Keyword', []))


def test_parser_noarg_trailingspc():
    check_config_parser('Keyword ', ('Keyword', []))


def test_parser_noarg_leadingspc():
    check_config_parser(' Keyword', ('Keyword', []))


def test_parser_onearg():
    check_config_parser('Keyword Arg1', ('Keyword', ['Arg1']))


def test_parser_onearg_dblspc():
    check_config_parser('Keyword  Arg1', ('Keyword', ['Arg1']))


def test_parser_twoargs():
    check_config_parser('Keyword Arg1 Arg2', ('Keyword', ['Arg1', 'Arg2']))


def test_parser_twoargs_dblspc():
    check_config_parser('Keyword  Arg1  Arg2', ('Keyword', ['Arg1', 'Arg2']))


def test_parser_quotearg_nospc():
    check_config_parser('Keyword "Arg1"', ('Keyword', ['Arg1']))


def test_parser_quotearg_withspc():
    check_config_parser('Keyword "Arg1 NotArg2"', ('Keyword', ['Arg1 NotArg2']))


def test_parser_quotearg_withspc_and_leading():
    check_config_parser('Keyword  "Arg1 NotArg2"', ('Keyword', ['Arg1 NotArg2']))


def test_parser_quotearg_withspc_and_trailing():
    check_config_parser('Keyword "Arg1 NotArg2" ', ('Keyword', ['Arg1 NotArg2']))


def test_parser_quotearg_withspc_and_leadin_and_trailing():
    check_config_parser('Keyword  "Arg1 NotArg2" ', ('Keyword', ['Arg1 NotArg2']))


def test_parser_quotearg_noquotearg():
    check_config_parser('Keyword "Arg1 NotArg2" Arg2', ('Keyword', ['Arg1 NotArg2', 'Arg2']))


def test_parser_noquotearg_quotearg():
    check_config_parser('Keyword Arg1 "Arg2 NotArg3"', ('Keyword', ['Arg1', 'Arg2 NotArg3']))


def test_parser_quotearg_noquotearg_quotearg():
    check_config_parser('Keyword "Arg1 NotArg2" Arg2', ('Keyword', ['Arg1 NotArg2', 'Arg2']))


def test_parser_equals_nospc():
    check_config_parser('Keyword=Value', ('Keyword', ['Value']))


def test_parser_equals_leftspc():
    check_config_parser('Keyword =Value', ('Keyword', ['Value']))


def test_parser_equals_rightspc():
    check_config_parser('Keyword= Value', ('Keyword', ['Value']))


def test_parser_equals_bothspc():
    check_config_parser('Keyword = Value', ('Keyword', ['Value']))


def test_parser_equals_allthespc():
    check_config_parser('Keyword   =   Value', ('Keyword', ['Value']))


def check_to_line(keyword, parts, expected, **kwargs):
    result = ConfigOutput.to_line(keyword, parts, **kwargs)
    eq_(result, expected)


def test_to_line_no_args():
    check_to_line('Test', [], 'Test')


def test_to_line_one_arg():
    check_to_line('Test', ['Hello'], 'Test = Hello')


def test_to_line_one_arg_with_quotes():
    check_to_line('Test', ['Hello "fun fun'], 'Test = Hello "fun fun')


def test_to_line_two_args():
    check_to_line('Test', ['Arg1', 'Arg2'], 'Test Arg1 Arg2')


def test_to_line_two_args_spaces():
    check_to_line('Test', ['Arg1 is Spacey', 'Arg2 is also Spacey'], 'Test "Arg1 is Spacey" "Arg2 is also Spacey"')


@raises(OutputException)
def test_to_line_two_args_spaces_quotes():
    ConfigOutput.to_line('Test', ['This has a quote"', 'Eep'])


@raises(ParserException)
def test_root_is_fails():
    config_for_text('@is a-thing')


@raises(ParserException)
def test_invalid_range_fails():
    Host.expand_with(['{1}'])


@raises(ParserException)
def test_invalid_range_dup_fails():
    Host.expand_with(['{1..2..4}'])


@raises(ParserException)
def test_invalid_range_nonint_fails():
    Host.expand_with(['{1..cat}'])


@raises(ParserException)
def test_invalid_range_empty():
    Host.expand_with(['{}'])


def test_expand():
    eq_(['1', '2', '3'], Host.expand_with(['{1..3}']))


def test_expand_range():
    eq_(['1', '3'], Host.expand_with(['{1..3/2}']))


@raises(SecurityException)
def test_http_disallowed():
    config_for_text("@include http://example.com/thing.sedge")


def test_subst():
    check_parse_result('@set goat cheese\n@set username percy\nHost <goat>\nUsername <username>', 'Host = cheese\n    Username = percy\n')


def test_subst_with_via():
    check_parse_result('@set goat cheese\n\nHost test\n@via <goat>', 'Host = test\n    ProxyCommand = ssh cheese nc %h %p 2> /dev/null\n')



from setuptools import setup, find_packages

dev_requires = ['flake8', 'nose']
install_requires = ['requests']

setup(
    name = "sedge",
    version = "1.1.0",
    packages = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    extras_require = {
        'dev': dev_requires
    },
    install_requires = install_requires,
    entry_points = {
        'console_scripts': [
            'sedge = sedge.cli:main',
        ],
    }
)
